/*
	Streams - Audio and Video Player Module
	Copyright 2004-2008 Ricky Romero and Harry Whitfield

	This program is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the
	Free Software Foundation; either version 2 of the License, or (at your
	option) any later version.

	This program is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; if not, write to the Free Software Foundation, Inc.,
	51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
	
	Streams - Audio and Video Player Module - for KonAVio Media 1.0b45
	4 March, 2008
	Copyright 2004-2008 Ricky Romero and Harry Whitfield
	mailto:g6auc@arrl.net
*/

/*
	Streams is a module that sits on top of the Media Player Classes and provides a more uniform
	and more convenient interface to the Media Player Classes for use by User Interfaces.

	The two principal functions provided are:
	
	function playStream(streamURL, volume, muted, paused, hidden, streamMode, linkName)
	
	and 
	
	function isStream(streamURL).
	
		streamURL: 	string 	- the url of the stream/file
		volume: 	number 	- 0..100
		muted: 		boolean
		paused: 	boolean
		hidden: 	boolean	- normally set to false
		
		streamMode: string	- "0","2","3","4","5"
			"Automatic":              streamMode = "0"
			"QuickTime Player":       streamMode = "2"
			"RealPlayer":             streamMode = "3"
			"Windows Media Player":   streamMode = "4"
			"Application or Browser": streamMode = "5"
		
		linkName:	string 	- the url of an image


	There are also some lower level functions which can be used, although these are not recommended:

	function playWMPstream(streamURL, volume, muted, paused, hidden),
	function playRPstream(streamURL, volume, muted, paused, hidden),
	function playQTstream(streamURL, volume, muted, paused, hidden) and
	function playSWFstream(streamURL, volume, muted, paused, hidden).
	
	
	The module has functions that may be called by User Interfaces to control the playing of media:

	function quitAllMediaPlayers(),
	function setMediaPlayerVolume(volume),
	function setMediaPlayerMuted(muted),
	function setMediaPlayerPaused(paused),
	
	function setFlashPlayerVolume(volume) and
	function setFlashPlayerMuted(muted).
	
	The last two actually control the system volume.
	
	
	There is also a function to reset the Media Types preference pane:
	
	function restoreDefaultExtns().
	
	
	The module makes the following callbacks on the User Interface:
	
	function setPlayerPlayButton(),
	function setPlayerPauseButton(),
	
	function setFramePlayButton() and 
	function setFramePauseButton().
	
	The last two are used with a Flash Player.
	
	If the User Interface does not already provide these functions, dummy functions should provided.
	
	
	This module may also call on the following functions to play Flash streams:
	
	function openFlash(streamURL, volume, muted, paused, hidden)
	
	and
	
	function quitFlash().
	
	Note that software to play Flash Movies is not provided by the Media Player classes.
*/

////////////////////////////////////////// Media Class Includes //////////////////////////////////////////

include("Resources/WMP.js");
include("Resources/RP.js");
if (system.platform == 'windows') { include('Resources/winQT.js'); }


///////////////////////////////////////////// Streams Globals ////////////////////////////////////////////

var allowHTTP = true;	// Allow HTTP to be a stream protocol - set false in WebImages/DropMover/DrawDemo

var wmpPlayer  = null;
var rpPlayer   = null;
var qtPlayer   = null;
var swfPlayer  = null;

var gVolume = 75;
var gVolumeMuted = false;

var gSystemVolume = Math.round(6.25*system.volume);
var gSystemVolumeMuted = system.muted;

var gMediaPaused = false;
var gMediaDuration = 0;
var gMediaDurationFormatted = 0;

var gFlashPaused = true;
var gFlashLoaded = 0;


/////////////////////////////////////////// Legacy Preferences ///////////////////////////////////////////

preferences.realAlternativePref.hidden = true;
preferences.realAlternativePref.value = "0";

preferences.convertBackSlashPref.hidden = true;
preferences.convertBackSlashPref.value = "1";

preferences.escapeURLPref.hidden = true;
preferences.escapeURLPref.value = "0";


/////////////////////////////////////////// Preference settings //////////////////////////////////////////

preferences.appHidePref.hidden = (systemPlatform == "windows");
if (systemPlatform == "windows") { preferences.appHidePref.value = "0"; }

var useRealAlternative = false;
var useFsOpen = (preferences.useFsOpenPref.value == "1");
var appHide   = (preferences.appHidePref.value   == "1");


////////////////////////////////////////// Lower Level Functions /////////////////////////////////////////

function playWMPstream(streamURL, volume, muted, paused, hidden)
{
	eprint('playWMPstream("' + streamURL + '", ' + volume + ', ' + muted + ', ' + paused + ', ' + hidden + ')');

	if (rpPlayer)  { rpPlayer.quit();  rpPlayer  = null; }
	if (qtPlayer)  { qtPlayer.quit();  qtPlayer  = null; }
	if (swfPlayer) { quitFlash();      swfPlayer = null; }
	if (wmpPlayer === null) { wmpPlayer = new WMP(streamURL); } else { wmpPlayer.setURL(streamURL); }
	wmpPlayer.setVolume(volume);
	wmpPlayer.setMute(muted);
	wmpPlayer.setMiniaturized(hidden);
	
	gMediaDuration = wmpPlayer.mediaDuration();
	eprint('gMediaDuration: ' + gMediaDuration);
	
	if (isNaN(gMediaDuration))  { gMediaDuration =  86397; }	// marker for NaN
	if (gMediaDuration < 0)     { gMediaDuration =  86398; }	// marker for Negative
	if (gMediaDuration > 86399) { gMediaDuration =  86399; } 	// Marker for too big
	
	gMediaDurationFormatted = WMP.formatTime(gMediaDuration);
	eprint('gMediaDurationFormatted: ' + gMediaDurationFormatted);

	if (paused) { wmpPlayer.pause(); } else { wmpPlayer.play(); }
	gMediaPaused = paused;
	if (paused) { setPlayerPlayButton(); } else { setPlayerPauseButton(); }			// *** callback to user interface
}

function playRPstream(streamURL, volume, muted, paused, hidden)
{
	if (useRealAlternative) { playWMPstream(streamURL, volume, muted, paused, hidden); return; }

	eprint('playRPstream("' + streamURL + '", ' + volume + ', ' + muted + ', ' + paused + ', ' + hidden + ')');

	if (wmpPlayer) { wmpPlayer.quit(); wmpPlayer = null; }
	if (qtPlayer)  { qtPlayer.quit();  qtPlayer  = null; }
	if (swfPlayer) { quitFlash();      swfPlayer = null; }
	if (rpPlayer === null) { rpPlayer = new RP(streamURL); } else { rpPlayer.setURL(streamURL); }
	rpPlayer.setVolume(volume);
	rpPlayer.setMute(muted);
	rpPlayer.setMiniaturized(hidden);

	if (paused) { rpPlayer.pause();  } else { rpPlayer.play();  }
	gMediaPaused = paused;
	if (paused) { setPlayerPlayButton(); } else { setPlayerPauseButton(); }			// *** callback to user interface
}

function playQTstream(streamURL, volume, muted, paused, hidden)
{
	if (systemPlatform == 'macintosh') { playWMPstream(streamURL, volume, muted, paused, hidden); return; }

	eprint('playQTstream("' + streamURL + '", ' + volume + ', ' + muted + ', ' + paused + ', ' + hidden + ')');

	if (wmpPlayer) { wmpPlayer.quit(); wmpPlayer = null; }
	if (rpPlayer)  { rpPlayer.quit();  rpPlayer  = null; }
	if (swfPlayer) { quitFlash();      swfPlayer = null; }
	if (qtPlayer === null) { qtPlayer = new QT(); qtTimer.ticking = true; }
	else
	{
		qtPlayer.setURL(streamURL);
		qtPlayer.setVolume(volume);
		qtPlayer.setMute(muted);
		qtPlayer.setMiniaturized(hidden);
		if (paused) { qtPlayer.pause(); } else { qtPlayer.play(); }
		gMediaPaused = paused;
		if (paused) { setPlayerPlayButton(); } else { setPlayerPauseButton(); }		// *** callback to user interface
	}
}

function startQT() { qtPlayer.setURL(qtPlayer.getURL()); }

function playSWFstream(streamURL, volume, muted, paused, hidden)
{
	eprint('playSWFstream("' + streamURL + '", ' + volume + ', ' + muted + ', ' + paused + ', ' + hidden + ')');

	if (qtPlayer)  { qtPlayer.quit();  qtPlayer  = null; }
	if (rpPlayer)  { rpPlayer.quit();  rpPlayer  = null; }
	if (wmpPlayer) { wmpPlayer.quit(); wmpPlayer = null; }
	gMediaPaused = true;
	setPlayerPlayButton();															// *** callback to user interface
	
	gFlashPaused = paused;
	openFlash(streamURL, volume, muted, paused, hidden);
	//if (paused) { setFramePlayButton(); } else { setFramePauseButton(); }			// *** callback to user interface
}


//////////////////////////////////////// User Interface Functions/////////////////////////////////////////

function quitAllMediaPlayers()
{
	if (wmpPlayer) { wmpPlayer.quit(); wmpPlayer = null; }
	if (rpPlayer)  { rpPlayer.quit();  rpPlayer  = null; }
	if (qtPlayer)  { qtPlayer.quit();  qtPlayer  = null; }
	if (swfPlayer) { quitFlash();      swfPlayer = null; }
	gMediaPaused = true;
	setPlayerPlayButton();															// *** callback to user interface
}

function setMediaPlayerVolume(volume)
{
	if (wmpPlayer) { wmpPlayer.setVolume(volume); }
	if (rpPlayer)  { rpPlayer.setVolume(volume);  }
	if (qtPlayer)  { qtPlayer.setVolume(volume);  }
	gVolume = volume;
}

function setMediaPlayerMuted(muted)
{
	if (wmpPlayer) { wmpPlayer.setMute(muted); }
	if (rpPlayer)  { rpPlayer.setMute(muted);  }
	if (qtPlayer)  { qtPlayer.setMute(muted);  }
	gVolumeMuted = muted;
}

function setMediaPlayerPaused(paused)
{
	if (wmpPlayer) { if (paused) { wmpPlayer.pause(); } else { wmpPlayer.play(); } gMediaPaused = paused; } else
	if (rpPlayer)  { if (paused) { rpPlayer.pause();  } else { rpPlayer.play();  } gMediaPaused = paused; } else
	if (qtPlayer)  { if (paused) { qtPlayer.pause();  } else { qtPlayer.play();  } gMediaPaused = paused; }
	else { beep(); return; }
}

function setFlashPlayerVolume(volume)
{
	system.volume = Math.round(0.16*volume);
	gSystemVolume = volume;
}

function setFlashPlayerMuted(muted)
{
	system.mute = muted;
	gSystemVolumeMuted = muted;
}


//////////////////////////////////////////// Reset Media Types ///////////////////////////////////////////

function restoreDefaultExtns()			// called by onPreferencesChanged handler to reset "Media Types" preferences
{
	if (preferences.restoreDefaultExtnsPref.value == "1")
	{
		preferences.restoreDefaultExtnsPref.value = "0";
		
		preferences.QTExtnPref.value  = preferences.QTExtnPref.defaultValue;
		preferences.RPExtnPref.value  = preferences.RPExtnPref.defaultValue;
		preferences.WMPExtnPref.value = preferences.WMPExtnPref.defaultValue;

		preferences.QTVideoExtnPref.value  = preferences.QTVideoExtnPref.defaultValue;
		preferences.RPVideoExtnPref.value  = preferences.RPVideoExtnPref.defaultValue;
		preferences.WMPVideoExtnPref.value = preferences.WMPVideoExtnPref.defaultValue;
		
		preferences.metaFilePref.value = preferences.metaFilePref.defaultValue;
	}
}


//////////////////////////////////////////// Private Functions ///////////////////////////////////////////

function isMediaStream(extn, player)
{
	var xtn = preferences[player].value.split(",");
	for (var i = 0; i < xtn.length; i += 1)
	{
		if ((xtn[i] !== "") && (("." + xtn[i]) == extn)) { return true; }
	}
	return false;
}

function isQTstream(extn)  { return isMediaStream(extn, "QTExtnPref"); }
function isRPstream(extn)  { return isMediaStream(extn, "RPExtnPref"); }
function isWMPstream(extn) { return isMediaStream(extn, "WMPExtnPref"); }

function isQTVideoStream(extn)  { return isMediaStream(extn, "QTVideoExtnPref"); }
function isRPVideoStream(extn)  { return isMediaStream(extn, "RPVideoExtnPref"); }
function isWMPVideoStream(extn) { return isMediaStream(extn, "WMPVideoExtnPref"); }

function isMetaFile(extn)
{
	var xtn = preferences.metaFilePref.value.split(",");
	for (var i = 0; i < xtn.length; i += 1)
	{
		if ((xtn[i] !== "") && (("." + xtn[i]) == extn)) { return true; }
	}
	return false;
}

function isStreamExtn(extn)
{
	return (extn == ".swf") || (extn == ".flv") || isQTstream(extn) || isRPstream(extn) || isWMPstream(extn) ||
	  isQTVideoStream(extn) || isRPVideoStream(extn) || isWMPVideoStream(extn) || isMetaFile(extn);
}

function isStreamProtocol(streamURL, allowHTTP)
{
	var protocol, regexpURL = /\s*(\w+:\/\/)?([\w\-\.\:]+)?(\/[^\s\?]*)?\??(\S*)/;
	var res = streamURL.match(regexpURL);
	if (res !== null)
	{
		protocol = res[1];
		return (protocol == "rtsp://") || (protocol == "mms://") || ((protocol == "http://") && allowHTTP);
	}
	return false;
}


/////////////////////////////////////////// Principal Functions //////////////////////////////////////////

function isStream(streamURL)
{
	function xtn(s) {
    	var idx = s.lastIndexOf("."), ext;
    	if (idx >= 0) {
       		 ext = s.substring(idx);
        	if (ext.length > 8) { return ""; }
        	if (ext.length > 1) { return ext; }
    	}
    	return "";
	}
	return isStreamProtocol(streamURL, allowHTTP) || isStreamExtn(xtn(streamURL));
}

function playStream(streamURL, volume, muted, paused, hidden, streamMode, linkName)
{
	function xtn(s) {
    	var idx = s.lastIndexOf("."), ext;
    	if (idx >= 0) {
       		 ext = s.substring(idx);
        	if (ext.length > 8) { return ""; }
        	if (ext.length > 1) { return ext; }
    	}
    	return "";
	}

	function report(streamURL, legend)
	{
		if (preferences.alertPref.value != "0")
		{
			beep();
		 	alert(legend + ": " + streamURL);
		   eprint(legend + ": " + streamURL);
		}
	}

	function isLocalFile(streamURL)
	{ 
		if (streamURL.indexOf('file:///') === 0) { return true; }
		if (((systemPlatform == "macintosh") && (streamURL[0] == '/')) ||
		((systemPlatform == "windows")   && (streamURL[1] == ':'))) { return true; }
		return false;
	}

	function openAsURL(streamURL)
	{
		quitAllMediaPlayers();
		if ((systemPlatform == "windows") && (streamURL.indexOf('file:///') === 0)) { streamURL = streamURL.substring(8); }
		eprint('openAsURL:openURL: ' + openURL(streamURL));
	}

	function playStreamProtocol(streamURL, volume, muted, paused, hidden, streamMode)
	{
		var protocol, regexpURL = /\s*(\w+:\/\/)?([\w\-\.\:]+)?(\/[^\s\?]*)?\??(\S*)/;
		var res = streamURL.match(regexpURL);
		if (res !== null)
		{
			protocol = res[1];
			switch (protocol)
			{
				case "rtsp://":	 playRPstream(streamURL, volume, muted, paused, hidden || appHide); return;
				case "mms://": 	playWMPstream(streamURL, volume, muted, paused, hidden || appHide); return;
				case "http://": 
					   			if (systemPlatform == "macintosh") { playRPstream(streamURL, volume, muted, paused); }
		   						else { playWMPstream(streamURL, volume, muted, paused); }
								return;
				default:		break;
			}
		}
		
		if (preferences.alertPref.value != "0") { report(streamURL, "playStreamProtocol:Could not get stream protocol for streamURL"); }
	}
	
	function playMetaFile(streamURL, volume, muted, paused, hidden, streamMode)
	{
		function playKnownStream(streamURL, volume, muted, paused, hidden, streamMode)
		{
			var extn = xtn(streamURL);
			if (isQTstream(extn))  		{  playQTstream(streamURL, volume, muted, paused, hidden || appHide); } else
			if (isQTVideoStream(extn))  {  playQTstream(streamURL, volume, muted, paused, hidden); } else
			if (isRPstream(extn))  		{  playRPstream(streamURL, volume, muted, paused, hidden || appHide); } else
			if (isRPVideoStream(extn))
			{
				if (systemPlatform == "macintosh")	{ playRPstream(streamURL, volume, muted, paused, hidden); } else { openAsURL(streamURL); }
			} else
			if (isWMPstream(extn))		{ playWMPstream(streamURL, volume, muted, paused, hidden || appHide); } else
			if (isWMPVideoStream(extn))
			{
				if (systemPlatform == "macintosh")	{ playWMPstream(streamURL, volume, muted, paused, hidden); } else { openAsURL(streamURL); }
			} else
			if (isStreamProtocol(streamURL, allowHTTP)) { playStreamProtocol(streamURL, volume, muted, paused, hidden, streamMode); } else
			if (preferences.alertPref.value != "0") { report(streamURL, "playKnownStream:Could not play streamURL"); }
		}
		
		function fetchMetaFile(url, outputFile)
		{
			var myUrl = new URL();
			myUrl.outputFile = outputFile;
			var result = myUrl.fetch(url);
			var resp = myUrl.response;
			eprint('fetchMetaFile:result:\n' + result);
			eprint('fetchMetaFile:resp:  ' + resp);
			if (resp == 200) { return ""; }
			else { return 'Cannot fetch "' + url + '". HTTP Error code ' + resp + '.'; }
	 	}
	
		function extractURL(metaData)	// for use with Windows Media Metafiles
		{
			var pattern1 = /<ref +href *\= *\"([^ \"]+?)[ \"].*?>/gi;
			var pattern2 = /<ref +href *\= *\"([^ \"]+?)[ \"].*?>/i;
	
			var url = [];
			var found1 = metaData.match(pattern1);		// get all the matches
			if (found1 !== null)
			{
				for (var i = 0; i < found1.length; i += 1)
				{
					url[i] = found1[i].match(pattern2)[1];
					if (url[i].indexOf("mms://") === 0) { return url[i]; }
				}
				return url[0];
			}
			if (preferences.alertPref.value != "0") { report(streamURL, "extractURL:Could not find reference to stream in meta-data"); }
			return "";
		}
	
		function extractURL1(data)
		{
			var lookFor1 = /^File1\=(.+?)$/m;
			var res = data.match(lookFor1);
			if (res !== null) { return res[1]; }
			return "";
		}
	
		function extractURL2(data)
		{
			var lookFor1 = /^<audio src\=\"([^\"]+?)\"/m;
			var res = data.match(lookFor1);
			if (res !== null) { return res[1]; }
			return "";
		}
	
		var streamURL0 = streamURL, path = streamURL, metaFile, metaData, extn = xtn(streamURL), result;
		if (path.indexOf('file:///') === 0) { path = path.substring(8); }
		if (isLocalFile(path))															// local meta-file
		{
			path = unescape(path);
			path = path.replace(/\\/g, "/");
			metaFile = path;
			metaData = filesystem.readFile(metaFile);
		}
		else																			// remote meta-file
		{
			metaFile = system.widgetDataFolder + "/Downloads/metaFile"+ extn;
			result = fetchMetaFile(streamURL, metaFile);
			if (result === "") { metaData = filesystem.readFile(metaFile); } else { metaData = null; }
		}
		eprint('playMetaFile:metaFile:'   + metaFile);
		eprint('playMetaFile:metaData:\n' + metaData);
		
		if (metaData)
		{
			if (useFsOpen)
			{
				quitAllMediaPlayers();
				eprint('playMetaFile:open metaFile 1: ' + filesystem.open(metaFile));	// open metafile
			}
			else if ( extn == ".pls" )
			{
				streamURL = extractURL1(metaData);
				if (streamURL === "") { report(streamURL0, "Could not find url in meta-data"); return; }
		   		playKnownStream(streamURL, volume, muted, paused, hidden, streamMode);
			}
			else if ( extn == ".smil" )
			{
				streamURL = extractURL2(metaData);
				if (streamURL === "") { report(streamURL0, "Could not find url in meta-data"); return; }
		   		playKnownStream(streamURL, volume, muted, paused, hidden, streamMode);
			}
			else if ( extn == ".wax" )
			{
				streamURL = extractURL(metaData);			// check for xml data and extract url
				if (streamURL === "") { report(streamURL0, "Could not find url in meta-data"); return; }
				playWMPstream(streamURL, volume, muted, paused, hidden || appHide);
			}
			else if (( extn == ".asx" ) || ( extn == ".wmx" ) || ( extn == ".wvx" ))
			{
				streamURL = extractURL(metaData);			// check for asx xml data and extract url
				if (streamURL === "") { report(streamURL0, "Could not find url in meta-data"); return; }
				playKnownStream(streamURL, volume, muted, paused, hidden);
			}
			else if ( extn == ".ram" )
			{
				streamURL = metaData.replace(/\r\n?|\n/g, "");
				playRPstream(streamURL, volume, muted, paused, hidden || appHide);
			}
			else
			{
				quitAllMediaPlayers();
				eprint('playMetaFile:open metaFile 2: ' + filesystem.open(metaFile));	// open metafile
			}
		}
		else if (preferences.alertPref.value != "0") { report(streamURL, "playMetaFile:Could not get meta-data for streamURL"); }
	}
	
	var path = streamURL;

	if (isLocalFile(path))
	{
		if (path.indexOf('file:///') === 0) { path = path.substring(8); }
		if (!filesystem.itemExists(path)) { report(path, "The file does not exist or has been moved"); return; }
	}

	if (linkName != "None") { displayLinkImage(linkName); }
	if ((streamMode != "0")	&& (streamMode != "1"))		// let the user choose
	{
		switch (streamMode)
		{
			case "2":  playQTstream(streamURL,   volume, muted, paused, hidden); break;
			case "3":  playRPstream(streamURL,   volume, muted, paused, hidden); break;
			case "4": playWMPstream(streamURL,   volume, muted, paused, hidden); break;
			case "5":     openAsURL(streamURL);  break;
		}
		return;
	}

	var extn = xtn(streamURL);
	if ((extn == ".swf") || (extn == ".flv")) { playSWFstream(streamURL, volume, muted, paused, hidden); } else
	if (isQTstream(extn))  		{  playQTstream(streamURL, volume, muted, paused, hidden || appHide); } else
	if (isQTVideoStream(extn))  {  playQTstream(streamURL, volume, muted, paused, hidden); } else
	if (isRPstream(extn))  		{  playRPstream(streamURL, volume, muted, paused, hidden || appHide); } else
	if (isRPVideoStream(extn))
	{
		if (systemPlatform == "macintosh")	{ playRPstream(streamURL, volume, muted, paused, hidden); } else { openAsURL(streamURL); }
	} else
	if (isWMPstream(extn))		{ playWMPstream(streamURL, volume, muted, paused, hidden || appHide); } else
	if (isWMPVideoStream(extn))
	{
		if (systemPlatform == "macintosh")	{ playWMPstream(streamURL, volume, muted, paused, hidden); } else { openAsURL(streamURL); }
	} else
	if (isMetaFile(extn))  {  playMetaFile(streamURL, volume, muted, paused, hidden, streamMode); } else
	if (isStreamProtocol(streamURL, allowHTTP)) { playStreamProtocol(streamURL, volume, muted, paused, hidden, streamMode); } else
	if (preferences.alertPref.value != "0") { report(path, "playStream:Could not play streamURL"); }
}

/////////////////////////////////////// End of Principal Functions ///////////////////////////////////////

streamsDefined = true;
